use std::{str::FromStr, usize};

use thiserror::Error;

#[derive(Debug, PartialEq, Eq, Clone)]
pub struct DealtHand {
    hand: String,
    bid: usize,
}

impl DealtHand {
    #[cfg(test)]
    pub fn just_label(hand: impl Into<String>) -> Self {
        let hand = hand.into();
        Self { hand, bid: 1 }
    }
    #[cfg(test)]
    pub fn new(hand: impl Into<String>, score: usize) -> Self {
        let hand = hand.into();
        Self { hand, bid: score }
    }

    pub fn hand(&self) -> &str {
        &self.hand
    }

    pub const fn bid(&self) -> usize {
        self.bid
    }
}

#[derive(Debug, Error, PartialEq, Eq)]
pub enum InvalidStrDealtHand {
    #[error("Dealt hand must consist of 2 elements")]
    Not2Elements,
    #[error("Score must be an unsigned number")]
    InvalidUnsignedNumberScore,
}

impl FromStr for DealtHand {
    type Err = InvalidStrDealtHand;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut in_two = s.split_whitespace();
        match (in_two.next(), in_two.next()) {
            (Some(hand), Some(score)) => {
                let score: usize = score
                    .parse()
                    .map_err(|_| InvalidStrDealtHand::InvalidUnsignedNumberScore)?;
                let hand = hand.to_string();
                let done = DealtHand { hand, bid: score };
                Ok(done)
            }
            _ => Err(InvalidStrDealtHand::Not2Elements),
        }
    }
}

#[cfg(test)]
mod testing {
    use crate::solutions::day7::dealt_hand::{DealtHand, InvalidStrDealtHand};

    #[test]
    fn parse_str() {
        fn assert_case(given: &str, expected: Result<DealtHand, InvalidStrDealtHand>) {
            let actual = given.parse();
            assert_eq!(expected, actual, "Input: {}", given);
        }

        fn happy_paths() {
            assert_case(
                "32T3K 765",
                Ok(DealtHand {
                    hand: "32T3K".to_string(),
                    bid: 765,
                }),
            );
            assert_case(
                "T55J5 684",
                Ok(DealtHand {
                    hand: "T55J5".to_string(),
                    bid: 684,
                }),
            );
            assert_case(
                "KK677 28",
                Ok(DealtHand {
                    hand: "KK677".to_string(),
                    bid: 28,
                }),
            );
            assert_case(
                "KTJJT 220",
                Ok(DealtHand {
                    hand: "KTJJT".to_string(),
                    bid: 220,
                }),
            );
            assert_case(
                "QQQJA 483",
                Ok(DealtHand {
                    hand: "QQQJA".to_string(),
                    bid: 483,
                }),
            );
        }

        fn error_paths() {
            assert_case(
                "xxxxx aaa",
                Err(InvalidStrDealtHand::InvalidUnsignedNumberScore),
            );
            assert_case("xxxxx", Err(InvalidStrDealtHand::Not2Elements));
        }

        happy_paths();
        error_paths();
    }
}
